;; Generated package description from plz.el  -*- no-byte-compile: t -*-
(define-package "plz" "0.2" "HTTP library" '((emacs "26.3")) :commit "9e9a370830a879592a4c6c9e66a5e9ec9f27d698" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("comm" "network" "http") :url "https://github.com/alphapapa/plz.el")
